package gov.usgs.earthquake.nshmp.gmm;

import static com.google.common.base.Preconditions.checkArgument;
import static gov.usgs.earthquake.nshmp.data.DoubleData.checkInRange;
import static java.lang.Math.abs;
import static java.lang.Math.log;
import static java.lang.Math.sqrt;

import java.util.EnumMap;
import java.util.List;
import java.util.Map;
import java.util.Optional;

import com.google.common.annotations.Beta;
import com.google.common.collect.FluentIterable;
import com.google.common.collect.Lists;
import com.google.common.collect.Range;
import com.google.common.primitives.Doubles;

import gov.usgs.earthquake.nshmp.data.Interpolator;
import gov.usgs.earthquake.nshmp.data.XySequence;
import gov.usgs.earthquake.nshmp.tree.LogicTree;
import gov.usgs.earthquake.nshmp.tree.Trees;

/**
 * USGS implementation of damping scaling factors (DSF).
 *
 * <p>Most GMMs for elastic response spectra are typically developed at a 5%
 * viscous damping ratio. This implementation uses the models of Rezaeian et al.
 * (2014) and Rezaeian et al. (2021) to compute DSFs that adjust damping ratios
 * to values between 0.5% and 30% for different tectonic settings.
 *
 * <p><b>Implementation notes:</b>
 *
 * <ul><li>Active crust model uses coefficients for RotD50, Table 1 in Rezaeian
 * et al. (2014)
 *
 * </ul>
 *
 * <p><b>Reference:</b> Rezaeian, S., Bozorgnia, Y., Idriss, I.M., Abrahamson,
 * N., Campbell, K., and Silva, W., 2014, Damping scaling factors for elastic
 * response spectra for shallow crustal earthquakes in active tectonic regions:
 * "Average" horizontal component: Earthquake Spectra, v. 30, n. 2, p. 939-963,
 * doi: <a href="http://dx.doi.org/10.1193/100512EQS298M">
 * 10.1193/100512EQS298M</a>.
 *
 * <p><b>Reference:</b> Rezaeian, S., Al Atik, L., Kuehn, N.M., Abrahamson, N.,
 * Bozorgnia, Y., Mazzoni, S., Withers, K., and Campbell, K., 2021, Spectral
 * damping scaling factors for horizontal components of ground motions from
 * subduction earthquakes using NGA-Subduction data: Earthquake Spectra, v. 37,
 * n. 4, p. 2453-2492, doi: <a href="https://doi.org/10.1177/87552930211027903">
 * 10.1177/87552930211027903</a>.
 *
 * @author U.S. Geological Survey
 */
@Beta
public class UsgsDampingScaling {

  /*
   * Developer notes:
   *
   * Current set up is to use instances by tectonic setting.
   *
   * The calcDsf() model returns a value much closer to zero for a damping ratio
   * of 5.027590868... The assumption is that only coarser damping ratios will
   * be used (e.g. 3.5, 4.0) such that this inconsistency will not make a
   * difference. Are we going to support a range of fixed values?
   *
   * For subduction, (1) special handling for periods > 5s, (2) correlation
   * coefficient table missing 5% column (is 0.0 in crustal), added column for
   * consistency, (3) What is done for rho below 0.1s period?
   *
   * Is linear interpolator appropriate for correlation coefficient
   * interpolation?
   *
   * TODO Disable subduction below 0.1s
   */

  /** The range of supported damping ratios: {@code [0.5..30]}. */
  public static final Range<Double> DAMPING_RATIO_RANGE = Range.closed(0.05, 30.0);

  private static final CoefficientContainer COEFFS_ACTIVE_CRUST;
  private static final CoefficientContainer COEFFS_INTERFACE;
  private static final CoefficientContainer COEFFS_SLAB;

  private static final String[] corrCoeffKeys =
      { "0.5", "1", "2", "3", "5", "7", "10", "15", "20", "25", "30" };

  private static final List<Double> CORR_COEFF_DSFS =
      FluentIterable.from(corrCoeffKeys)
          .transform(Doubles.stringConverter())
          .toList();

  private static final Interpolator CC_INTERP = Interpolator.builder().build();

  static {
    COEFFS_ACTIVE_CRUST = new CoefficientContainer("Rezaeian14-active-crust.csv");
    COEFFS_INTERFACE = new CoefficientContainer("Rezaeian21-interface.csv");
    COEFFS_SLAB = new CoefficientContainer("Rezaeian21-slab.csv");
  }

  private static final class Coefficients {

    final double a0, a1, b0, b1, b2, b3, b4, b5, b6, b7, b8;

    Coefficients(Imt imt, CoefficientContainer cc) {
      Map<String, Double> coeffs = cc.get(imt);
      a0 = coeffs.get("a0");
      a1 = coeffs.get("a1");
      b0 = coeffs.get("b0");
      b1 = coeffs.get("b1");
      b2 = coeffs.get("b2");
      b3 = coeffs.get("b3");
      b4 = coeffs.get("b4");
      b5 = coeffs.get("b5");
      b6 = coeffs.get("b6");
      b7 = coeffs.get("b7");
      b8 = coeffs.get("b8");
    }
  }

  private final Map<Imt, Coefficients> coeffs = new EnumMap<>(Imt.class);
  private final Map<Imt, XySequence> corrCoeffs = new EnumMap<>(Imt.class);

  // private final SourceType type;

  /**
   * Create a new damping scaling factor (DSF) model instance for the specified
   * tectonic settign and source type.
   *
   * @param setting tectonic setting for the DSF
   * @param type source type for the DSF; this argument is only relevant if
   *        {@code TectonicSetting == SUBDUCTION}
   */
  private UsgsDampingScaling(Gmm.Type type) {
    // Dallin- Modified the function to only use source type to determine the
    // correct coeffs
    // this.type = type;
    CoefficientContainer cc = (type == Gmm.Type.SUBDUCTION_INTERFACE)
        ? COEFFS_INTERFACE
        : (type == Gmm.Type.SUBDUCTION_SLAB)
            ? COEFFS_SLAB
            : COEFFS_ACTIVE_CRUST;

    // COEFFS_ACTIVE_CRUST;
    // cc = ()
    // if (type == Gmm.Type.SUBDUCTION_INTERFACE)
    // if (type == INTERFACE || type == INTERFACE_CLUSTER || type ==
    // INTERFACE_SYSTEM ||
    // type == INTERFACE_GRID) {
    // cc = COEFFS_INTERFACE;
    // } else if (type == SLAB || type == SLAB_GRID) {
    // cc = COEFFS_SLAB;
    // } else {
    // cc = COEFFS_ACTIVE_CRUST;
    // }

    for (Imt imt : cc.imts()) {
      coeffs.put(imt, new Coefficients(imt, cc));
      corrCoeffs.put(imt, initCorrCoeff(imt, cc));
    }
  }

  /*
   * Correlation coeffs at fixed damping ratios as a sequence for interpolation
   * over the range [0.5..30.0]
   */
  private static XySequence initCorrCoeff(Imt imt, CoefficientContainer cc) {
    Map<String, Double> coeffs = cc.get(imt);
    List<Double> corrCoeffs = Lists.newArrayList();
    for (String key : corrCoeffKeys) {
      corrCoeffs.add(coeffs.get(key));
    }
    return XySequence.create(CORR_COEFF_DSFS, corrCoeffs);
  }

  /**
   * Compute the damping scaling factor (DSF) in natural log units for the
   * damping ratio of interest.
   *
   * @param imt of interest
   * @param Mw magnitude of interest
   * @param rRup rupture distance of interest
   * @param dampingRatio in range [0.5..30]%
   */
  public double factor(Imt imt, double Mw, double rRup, double dampingRatio) {
    checkArgument(coeffs.containsKey(imt), "Unsupported IMT: %s", imt.name());
    checkInRange(DAMPING_RATIO_RANGE, "Damping ratio", dampingRatio);
    return (dampingRatio != 5.0) ? calcFactor(imt, Mw, rRup, dampingRatio) : 0;
    // double lnDSF = 0.0;
    // if (dampingRatio != 5.0) {
    // if (imt.period() > 5.0 &&
    // (type == INTERFACE || type == INTERFACE_CLUSTER || type ==
    // INTERFACE_SYSTEM ||
    // type == INTERFACE_GRID || type == SLAB || type == SLAB_GRID)) {
    // // System.out.println("LONGPERIOD");
    // lnDSF = calclnDsflongPeriod(imt, Mw, rRup, dampingRatio);
    // } else {
    // // System.out.println("normal");
    // lnDSF = calclnDsf(imt, Mw, rRup, dampingRatio);
    // }
    // }
    // return lnDSF;
  }

  double calcFactor(Imt imt, double Mw, double rRup, double dampingRatio) {
    return (imt.isSA() && imt.period() > 5.0)
        ? calclnDsflongPeriod(imt, Mw, rRup, dampingRatio)
        : calclnDsf(imt, Mw, rRup, dampingRatio);
  }

  /* Rezaeian et al. (2014, 2021) equation 3. */
  private double calclnDsf(
      Imt imt,
      double Mw,
      double rRup,
      double dampingRatio) {

    Coefficients c = coeffs.get(imt);
    double lnβ = log(dampingRatio);
    double lnβsq = lnβ * lnβ;
    return c.b0 + c.b1 * lnβ + c.b2 * lnβsq +
        (c.b3 + c.b4 * lnβ + c.b5 * lnβsq) * Mw +
        (c.b6 + c.b7 * lnβ + c.b8 * lnβsq) * log(rRup + 1.0);
  }

  /* Rezaeian et al. (2021) equation 5. */
  private double calclnDsflongPeriod(
      Imt imt,
      double Mw,
      double rRup,
      double dampingRatio) {

    double d0 = 0.6592 * Math.min(Mw, 8.5) - 5.3705;
    double d1 = 1 / (1 - Math.exp(4.3 * d0));
    double lnDSF5s = calclnDsf(Imt.SA5P0, Mw, rRup, dampingRatio);
    double lnDSFratio = 1 + d1 * Math.exp(d0 * Math.log(imt.period() / 5)) - d1;

    return lnDSFratio * lnDSF5s;
  }

  /**
   * Compute logarithmic standard deviation for the damping ratio of interest.
   *
   * @param imt of interest
   * @param dampingRatio in range [0.5..30]%
   */
  public double sigma(Imt imt, double dampingRatio) {
    checkArgument(coeffs.containsKey(imt), "Unsupported IMT: %s", imt.name());
    checkInRange(DAMPING_RATIO_RANGE, "Damping ratio", dampingRatio);
    return (dampingRatio == 5.0) ? 0.0 : calcSigma(imt, dampingRatio);
  }

  /* Rezaeian et al. (2014, 2021) equation 4. */
  private double calcSigma(Imt imt, double dampingRatio) {
    Coefficients c = coeffs.get(imt);
    double βTerm = log(dampingRatio / 5.0);
    return abs(c.a0 * βTerm + c.a1 * βTerm * βTerm);
  }

  /**
   * Compute the correlation coefficient between ln(DSF) and
   * ln(PSA<sub>5%</sub>).
   *
   * @param imt of interest
   * @param dampingRatio in range [0.5..30]%
   */
  public double rho(Imt imt, double dampingRatio) {
    checkArgument(coeffs.containsKey(imt), "Unsupported IMT: %s", imt.name());
    return calcRho(imt, dampingRatio);
  }

  /* Derived from tables in publications. */
  private double calcRho(Imt imt, double dampingRatio) {
    XySequence cc = corrCoeffs.get(imt);
    return CC_INTERP.findY(cc, dampingRatio);
  }

  /* Currently unused; equation for adjusting PSA sigma. */
  private static double updateSigma(double σPsa, double σDsf, double ρ) {
    return sqrt(σPsa * σPsa + σDsf * σDsf + 2.0 * σPsa * σDsf * ρ);
  }

  private static final UsgsDampingScaling INSTANCE_DEFAULT =
      new UsgsDampingScaling(Gmm.Type.ACTIVE_CRUST);
  private static final UsgsDampingScaling INSTANCE_INTERFACE =
      new UsgsDampingScaling(Gmm.Type.SUBDUCTION_INTERFACE);
  private static final UsgsDampingScaling INSTANCE_SLAB =
      new UsgsDampingScaling(Gmm.Type.SUBDUCTION_SLAB);

  public static LogicTree<GroundMotion> apply(
      Imt imt,
      GmmInput in,
      LogicTree<GroundMotion> tree,
      Gmm.Type type,
      double dampingRatio) {

    if (dampingRatio == 5.0) {
      return tree;
    }

    /*
     * Changed method signature
     *
     * Switch to set instance to use.
     *
     * factor() sigma() and rho() don't depend on mean and sigma values; only be
     * called once
     */

    UsgsDampingScaling instance = (type == Gmm.Type.SUBDUCTION_INTERFACE)
        ? INSTANCE_INTERFACE
        : (type == Gmm.Type.SUBDUCTION_SLAB)
            ? INSTANCE_SLAB
            : INSTANCE_DEFAULT;

    double factor = instance.calcFactor(imt, in.Mw, in.rRup, dampingRatio);
    double sigma = instance.calcSigma(imt, dampingRatio);
    double rho = instance.rho(imt, dampingRatio);

    return Trees.transform(
        tree,
        gm -> GroundMotion.create(
            gm.mean() + factor,
            updateSigma(gm.sigma(), sigma, rho)),
        Optional.empty());

    // if (type == Gmm.Type.SUBDUCTION_INTERFACE) {
    //
    // // if (gmmInput.srcType == INTERFACE || gmmInput.srcType ==
    // // INTERFACE_CLUSTER ||
    // // gmmInput.srcType == INTERFACE_SYSTEM || gmmInput.srcType ==
    // // INTERFACE_GRID) {
    //
    // return Trees.transform(tree,
    // gm -> GroundMotion.create(
    // gm.mean() +
    // INSTANCE_INTERFACE.factor(imt, gmmInput.Mw, gmmInput.rRup, dampingRatio),
    // updateSigma(gm.sigma(), INSTANCE_INTERFACE.sigma(imt, dampingRatio),
    // INSTANCE_INTERFACE.rho(imt, dampingRatio))),
    // Optional.empty());
    //
    // } else if (type == Gmm.Type.SUBDUCTION_SLAB) {
    // // if (gmmInput.srcType == SLAB || gmmInput.srcType == SLAB_GRID) {
    // return Trees.transform(tree,
    // gm -> GroundMotion.create(
    // gm.mean() +
    // INSTANCE_SLAB.factor(imt, gmmInput.Mw, gmmInput.rRup, dampingRatio),
    // updateSigma(gm.sigma(), INSTANCE_SLAB.sigma(imt, dampingRatio),
    // INSTANCE_SLAB.rho(imt, dampingRatio))),
    // Optional.empty());
    //
    // } else {
    // // active and stable crust
    // return Trees.transform(tree,
    // gm -> GroundMotion.create(
    // gm.mean() +
    // INSTANCE_DEFAULT.factor(imt, gmmInput.Mw, gmmInput.rRup, dampingRatio),
    // updateSigma(gm.sigma(), INSTANCE_DEFAULT.sigma(imt, dampingRatio),
    // INSTANCE_DEFAULT.rho(imt, dampingRatio))),
    // Optional.empty());
    // }
  }

  public static void main(String[] args) {

    UsgsDampingScaling model = new UsgsDampingScaling(Gmm.Type.SUBDUCTION_INTERFACE);

    Imt imt = Imt.SA5P0;
    double Mw = 7.0;
    double rRup = 120;
    double ratio = 5.0;

    double scaleFactor = Math.exp(model.factor(imt, Mw, rRup, ratio));
    double dsfSigma = model.sigma(imt, ratio);
    double rho = model.rho(imt, ratio);

    // System.out.println("Type: " + model.type);
    System.out.println("IMT: " + imt.name() + " Mw: " + Mw + " rRup: " + rRup +
        " Damping Ratio: " + ratio + "%");
    System.out.println("DSF: " + scaleFactor + " sigma: " + dsfSigma + " rho:" + rho);

  }
}
